/*
#    xfrmcalc - Scientific calculator based on Forms.
#    Copyright (C) 1996-1997  Martin Bartlett
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "forms.h"
#include "math.h"
#include "calculator.h"
#include "calchelp.h"
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>


#ifndef __GNU_LIBRARY__

/*
 * XFormCalc is written with the GNU long versions of math functions. If you
 * are not using the GNU library then these are macro'd to the standard ones.
 */

#define fabsl		fabs
#define sinl		sin
#define cosl		cos
#define tanl		tan
#define asinl		asin
#define acosl		acos
#define atanl		atan
#define sinhl		sinh
#define coshl		cosh
#define tanhl		tanh
#define asinhl		asinh
#define acoshl		acosh
#define atanhl		atanh
#define log10l		log10
#define logl		log
#define powl		pow
#define expl		exp
#define sqrtl		sqrt
#define fmodl		fmod

long double modfl(long double value, long double *ip)
{
	double		work1, work2;
	work2 = modf(value, &work1);
	*ip = work1;
	return work2;
}
	
#endif

#define PRECEDENCE_INCR	20

#define FUNC_NULL	0
#define FUNC_OR		1
#define FUNC_XOR	2
#define FUNC_AND	3
#define FUNC_LSH	4
#define FUNC_RSH	5
#define FUNC_ADD	6
#define FUNC_SUBTRACT	7
#define FUNC_MULTIPLY	8
#define FUNC_DIVIDE	9
#define FUNC_MOD	10
#define FUNC_POWER	11
#define FUNC_PWR_ROOT	12
#define FUNC_INTDIV	13
#define FUNC_COMMA	14
#define FUNC_COUNT  15 			/* Max number of functions
char	*function_desc[FUNC_COUNT] = {

	"Null",
	"Or",
	"Exclusive Or",
	"And",
	"Left Shift",
	"Right Shift",
	"Add",
	"Subtract",
	"Multiply",
	"Divide",
	"Modulus",
	"Power",
	"Reciprocal Power",
	"Integer Division",
	"Argument Stacking"
};

/*
 * Natural inverses/associations allow one calculator button to
 * have two different operations depending on which mouse button
 * is used. Short cut versions are implemented via hidden buttons.
 */
int	adjust_op[FUNC_COUNT][3] = {
				{FUNC_NULL,     FUNC_NULL,     FUNC_NULL},
				{FUNC_OR,       FUNC_OR,       FUNC_XOR },
				{FUNC_XOR,      FUNC_XOR,      FUNC_XOR },
				{FUNC_AND,      FUNC_AND,      FUNC_AND },
				{FUNC_LSH,      FUNC_LSH,      FUNC_RSH },
				{FUNC_RSH,      FUNC_RSH,      FUNC_RSH },
				{FUNC_ADD,      FUNC_ADD,      FUNC_ADD },
				{FUNC_SUBTRACT, FUNC_SUBTRACT, FUNC_SUBTRACT},
				{FUNC_MULTIPLY, FUNC_MULTIPLY, FUNC_MULTIPLY},
				{FUNC_DIVIDE,   FUNC_DIVIDE,   FUNC_DIVIDE},
				{FUNC_MOD,      FUNC_MOD,      FUNC_INTDIV },
				{FUNC_POWER,    FUNC_POWER,    FUNC_PWR_ROOT},
				{FUNC_PWR_ROOT, FUNC_PWR_ROOT, FUNC_PWR_ROOT},
				{FUNC_INTDIV,   FUNC_INTDIV,   FUNC_INTDIV},
				{FUNC_COMMA,    FUNC_COMMA,    FUNC_COMMA},
};

int	precedence[FUNC_COUNT] = { 0, 1, 2, 3, 4, 4, 5, 5, 6, 6, 6, 7, 7, 6, 8 };

CALCAMNT ExecOr(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecXor(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecAnd(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecLsh(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecRsh(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecAdd(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecSubtract(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecMultiply(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecComma(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecDivide(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecMod(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecPower(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecPwrRoot(CALCAMNT left_op, CALCAMNT right_op);
CALCAMNT ExecIntDiv(CALCAMNT left_op, CALCAMNT right_op);

CALCAMNT ExecAddSubP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result);
CALCAMNT ExecMultiplyP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result);
CALCAMNT ExecDivideP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result);
CALCAMNT ExecPowerP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result);
CALCAMNT ExecPwrRootP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result);

typedef	CALCAMNT  (*Arith)(CALCAMNT, CALCAMNT); 
typedef	CALCAMNT  (*Prcnt)(CALCAMNT, CALCAMNT, CALCAMNT); 
Arith		Arith_ops[FUNC_COUNT] = { NULL,
				ExecOr, 
				ExecXor,  
				ExecAnd,  
				ExecLsh,  
				ExecRsh,  
				ExecAdd, ExecSubtract, 
				ExecMultiply, ExecDivide, ExecMod,
				ExecPower, ExecPwrRoot, 
				ExecIntDiv, 
				ExecComma
};
Prcnt		Prcnt_ops[FUNC_COUNT] = { NULL,
				NULL, 
				NULL,
				NULL,
				NULL,
				NULL,
				ExecAddSubP,  ExecAddSubP,  
				ExecMultiplyP, ExecDivideP, ExecDivideP,
				ExecPowerP, ExecPwrRootP, 
				ExecDivideP, 
				NULL
};

typedef enum   _num_base	{ 
	NB_BINARY = 2, NB_OCTAL = 8, NB_DECIMAL = 10, NB_HEX = 16 
} num_base;

typedef enum   _angle_type	{ 
	ANG_DEGREE = 0, ANG_RADIAN = 1, ANG_GRADIENT = 2
} angle_type;

typedef enum   _item_type	{ 
	ITEM_FUNCTION, ITEM_AMOUNT 
} item_type;

typedef struct   _func_data	{ 
	int		item_function;
	int		item_precedence;
} func_data;

typedef	union  _item_data 	{	/* The item data	 */
	CALCAMNT	item_amount;	/*	an amount	 */
	func_data	item_func_data;	/*	or a function	 */
} item_data;				/* called item_data      */

typedef struct _item_contents	{	/* The item contents	 */
	item_type	s_item_type;	/* 	a type flag      */
	item_data	s_item_data; 	/*      and data	 */
} item_contents;

int			precedence_base = 0;
num_base	current_base = NB_DECIMAL;
int			overlay_function = 0;	/* Flag indicating whether to overlay */
int			input_limit = 0;
int			input_count = 0;

int		temp_button_numb = 0;

#define		DEC_SIZE	32
#define		BOH_SIZE	32
#define		HEX_SIZE	BOH_SIZE
#define		OCT_SIZE	BOH_SIZE
#define		BIN_SIZE	BOH_SIZE
#define		DSP_SIZE	33
item_contents   display_data = { ITEM_AMOUNT, 0 };
int		display_size = DEC_SIZE;
#define		DISPLAY_AMOUNT	display_data.s_item_data.item_amount

typedef struct _old_disp {   /* display statuc for backspace processing */
	CALCAMNT amount;
	int		 decimal;
} old_disp;

old_disp old_display[DSP_SIZE] = {{0,0}};
int		 old_display_index = 0;
int		refresh_display;
int		display_editable;
int		display_error = 0;
int		decimal_point = 0;
int		percent_mode = 0;
int		fl_sel_frozen;

/*
 * Trig stuff: we keep track of what hyperbolic mode we are in and what
 * angle representation the user needs, we have macros to help with
 * conversions as needed, and we have two arrays of pointers to
 * the real trig functions that we use, one for Mouse Button 1 (the
 * plain trig functions) and one for Mouse Button 3 (for the arcs).
 * We also frig 'close to zero' results!!!!
 *
 * We also do our own calculation of PI since pi is not accurate
 * enough (as evidenced by its effect on COS 90 degrees)!!!!
 */

int		hyp_mode = 0;
angle_type	angle_mode = ANG_DEGREE;
CALCAMNT	pi;

#define		DEG2RAD(x)	(((2L*pi)/360)*x)
#define		GRA2RAD(x)	((pi/200L)*x)
#define		RAD2DEG(x)	((360/(2L*pi))*x)
#define		RAD2GRA(x)	((200L/pi)*x)
#define		POS_ZERO	 1e-19	 /* What we consider zero is   */
#define		NEG_ZERO	-1e-19	 /* anything between these two */

typedef	CALCAMNT  (*Trig)(CALCAMNT); 
Trig		Trig_1[][3] = { 
		{  sinl,  cosl,  tanl },
		{ sinhl, coshl, tanhl }
};
Trig		Trig_3[][3] = { 
		{ asinl,  acosl,  atanl },
		{asinhl, acoshl, atanhl }
};

void ExecTrig(FL_OBJECT *ob, long data);
void ExecATrig(FL_OBJECT *ob, long data);

void 		UpdateDisplay(void);
int 		UpdateStack(int run_precedence);
int		 	ExecFFunction(int findex);
int		 	ExecVFunction(int findex);
int 		Exec01Function(int findex);
int			RunFunction(char *, int, char **, CALCAMNT *);
int 		cvb(char *out_str, long amount, int max_out);

void 		PrintStack(void);
void 		InitStack(void);
void 		PushStack(item_contents *add_item);
item_contents 	*PopStack(void);
item_contents 	*TopOfStack(void);
item_contents 	*TopTypeStack(item_type rqstd_type);

int 	LoseSel(FL_OBJECT *ob, long type);
int 	GainSel(FL_OBJECT *ob, long type, const void *data, long size);
void 	GainTimeout(int id, void *data);

void RefreshCalculator(void);

/*
 * Get around fl_get_button_numb bug
 */
int temp_get_button_numb(FL_OBJECT *ob)
{
	int num = fl_get_button_numb(ob);
	return (num ? num : temp_button_numb);
}

/* 
 * callbacks for form calculator 
 */

void InitializeCalculator(void) {

	/*
	 * Basic initialization involves initializing the calcultion,
	 * stack, forcing the display to refresh to zero, and setting
	 * up the floating point excetion signal handler to trap the
	 * errors that the code can/has not been written to trap.
	 *
	 * We also calculate pi as double the arc sine of 1. 
	*/

	void fpe_handler(int fpe_parm);
	struct	sigaction	fpe_trap;
	
	RefreshCalculator();

	fpe_trap.sa_handler = &fpe_handler;
	fpe_trap.sa_flags = SA_RESTART;

	sigaction(SIGFPE, &fpe_trap, NULL);

	pi = asinl(1) * 2;
}		

void fpe_handler(int fpe_parm)
{
	display_error = 1;
	DISPLAY_AMOUNT = 0;
 }		

void RefreshCalculator(void)
{
	InitStack();
	display_error = 0;
	DISPLAY_AMOUNT = 0;
	UpdateDisplay();
}		

void EnterDigit(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	if (refresh_display) {
		DISPLAY_AMOUNT = 0;
		old_display_index = 0;
		decimal_point = 0;
		refresh_display = 0;
		input_count = 0;
	}

	if (!(input_limit && input_count >= input_limit))
	{
		old_display[old_display_index].amount  = DISPLAY_AMOUNT;
		old_display[old_display_index++].decimal = decimal_point;
		if (DISPLAY_AMOUNT < 0)
			DISPLAY_AMOUNT = decimal_point ? 
				DISPLAY_AMOUNT - ((CALCAMNT)data / 
					pow(current_base, decimal_point++)) :
				(current_base * DISPLAY_AMOUNT) - data;
		else
			DISPLAY_AMOUNT = decimal_point ? 
				DISPLAY_AMOUNT + ((CALCAMNT)data / 
					pow(current_base, decimal_point++)) :
				(current_base * DISPLAY_AMOUNT) + data;
	}

	overlay_function = 0;
	UpdateDisplay();
}

void EnterDecimal(FL_OBJECT *ob, long data)
{
	int button = temp_get_button_numb(ob);
	if (display_error)
		return;

	if ((button < FL_SHORTCUT && button > 1) || button == FL_SHORTCUT+',')
	{
		EnterComma(ob, FUNC_COMMA);
		return;
	}

	overlay_function = 0;
	decimal_point = 1;
	if (refresh_display) {
		DISPLAY_AMOUNT = 0;
		old_display_index = 0;
		refresh_display = 0;
		input_count = 1;
		UpdateDisplay();
	}
}

void EnterFunction(FL_OBJECT *ob, long data)
{
	int func = fl_get_menu(ob)-1;

	if (display_error)
		return;

	/*
	 * The rules for stack manipulation are quite complex
	 * based upon the number of arguments taken by the function.
	 * Therefore, the three major cases are handled by three 
	 * different routines.
	 */
	if (ext_funcs[func].args == 0 || ext_funcs[func].args == 1)
		Exec01Function(func);
	else if (ext_funcs[func].args < 0)
		ExecVFunction(func);
	else
		ExecFFunction(func);
	
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}
	
void EnterComma(FL_OBJECT *ob, long data)
{
	item_contents 	new_item, *top_item;
	int		new_precedence;
/*
 * The comma acts like a two operand function - almost! It can be used
 * to override a previous input, but its main purpose is to stack arguments
 * for a function. Therefore we don't update the stack when we have a comma,
 * but we add to the stack so that we can roll the stack to the end of
 * the arguments when we need to, or do the overrides when no external
 * function is executed.
 */ 
	if (display_error)
		return;

	if (!overlay_function)
		PushStack(&display_data);

	/*
	 * If the top item is a function, we replace it with
	 * our own function
	 */
	if ((top_item = TopOfStack())
	&& (top_item->s_item_type == ITEM_FUNCTION))
		PopStack();

	new_item.s_item_type = ITEM_FUNCTION;
	new_item.s_item_data.item_func_data.item_function = data;
	new_item.s_item_data.item_func_data.item_precedence =
		new_precedence = precedence[data] + precedence_base;

	overlay_function = 1;
	refresh_display = 1;
	PushStack(&new_item);
}

void EnterStackFunction(FL_OBJECT *ob, long data)
{
	item_contents 	new_item, *top_item;
	int		new_precedence, button;

	/*
     * First adjust the function number based upon which mouse
	 * button was hit
	 */

	if (display_error)
		return;

	if ((button = temp_get_button_numb(ob)) < FL_SHORTCUT)
		data = adjust_op[data][button];

	/*
	 * If something else has happened since we were last here
	 * then we push the display value
	 */
	if (!overlay_function)
		PushStack(&display_data);

	/*
	 * If the top iyem is a function, we replace it with
	 * our own function
	 */
	if ((top_item = TopOfStack())
	&& (top_item->s_item_type == ITEM_FUNCTION))
		PopStack();

	new_item.s_item_type = ITEM_FUNCTION;
	new_item.s_item_data.item_func_data.item_function = data;
	new_item.s_item_data.item_func_data.item_precedence =
		new_precedence = precedence[data] + precedence_base;

	overlay_function = 1;
	refresh_display = 1;
	if (UpdateStack(new_precedence))
		UpdateDisplay();
	PushStack(&new_item);
}

void EnterNegate(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	overlay_function = 0;
	if (DISPLAY_AMOUNT != 0) {
		DISPLAY_AMOUNT *= -1;
		UpdateDisplay();
	}
}

void EnterOpenParen(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	precedence_base += PRECEDENCE_INCR;
	overlay_function = 0;
	refresh_display = 1;
}

void EnterCloseParen(FL_OBJECT *ob, long data)
{
	item_contents 	*top_item;
	if (display_error)
		return;

	if (!overlay_function)
		PushStack(&display_data);
	/*
	 * If the top item is a function, we discard it
	 */

	if ((top_item = TopOfStack())
	&& (top_item->s_item_type == ITEM_FUNCTION))
		PopStack();

	overlay_function = 0;
	refresh_display = 1;
	if (UpdateStack(precedence_base))
		UpdateDisplay();
	if ((precedence_base -= PRECEDENCE_INCR) < 0)
		precedence_base = 0;
}

void EnterRecip(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	if (DISPLAY_AMOUNT == 0)
		display_error = 1;
	else
		DISPLAY_AMOUNT = 1 / DISPLAY_AMOUNT;	
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void EnterInt(FL_OBJECT *ob, long data)
{
	CALCAMNT work_amount1, work_amount2;
	int button;

	if (display_error)
		return;

	work_amount2 = modfl(DISPLAY_AMOUNT, &work_amount1);

	if ((button = temp_get_button_numb(ob)) < FL_SHORTCUT)
		data = (button > 2);

	DISPLAY_AMOUNT =  data ? work_amount2 : work_amount1;
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void EnterFactorial(FL_OBJECT *ob, long data)
{
	CALCAMNT work_amount1, work_amount2;
	int	 incr;

	if (display_error)
		return;

	modfl(DISPLAY_AMOUNT, &work_amount1);

	incr = work_amount1 < 0 ? -1 : 1;
	work_amount2 = work_amount1 - incr;
	while (work_amount1 != 0 && work_amount2 != 0 && !display_error) {
		work_amount1 *= work_amount2;
		work_amount2 -= incr;
	}

	DISPLAY_AMOUNT = work_amount1;
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void EnterSquare(FL_OBJECT *ob, long data)
{
	int button;

	if (display_error)
		return;

	if ((button = temp_get_button_numb(ob)) < FL_SHORTCUT)
		data = (button > 2);

	if (!data) {
		DISPLAY_AMOUNT *= DISPLAY_AMOUNT;	
	} else if (DISPLAY_AMOUNT < 0)
		display_error = 1;
	else
		DISPLAY_AMOUNT = sqrtl(DISPLAY_AMOUNT);	
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void EnterNotCmp(FL_OBJECT *ob, long data)
{
	CALCAMNT	boh_work_d;
	long 		boh_work;

	if (display_error)
		return;

	modfl(DISPLAY_AMOUNT, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX)
		display_error = 1;    
	else {
		boh_work = boh_work_d;
		DISPLAY_AMOUNT = ~boh_work;	
	} 
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void EnterHyper(FL_OBJECT *ob, long data)
{
	/*
	 * Set/Reset Hyperbolic modes of trig functions
	 */
	hyp_mode = !hyp_mode;
}

void EnterTrig(FL_OBJECT *ob, long data)
{
	int button = temp_get_button_numb(ob);

	if (button > FL_SHORTCUT)
		button = (button > (FL_SHORTCUT + XK_F12)) ? 3 : 1;

	if ((button = temp_get_button_numb(ob)) < 3) 
		ExecTrig(ob, data);
	else 
		ExecATrig(ob, data);
	overlay_function = 0;
}

void ExecTrig(FL_OBJECT *ob, long data)
{
	CALCAMNT 	work_amount;

	if (display_error)
		return;

	if (hyp_mode)
		work_amount = DISPLAY_AMOUNT;
	else {
		switch (angle_mode) {
			case ANG_DEGREE:
				work_amount = DEG2RAD(DISPLAY_AMOUNT);
				break;
			case ANG_GRADIENT:
				work_amount = GRA2RAD(DISPLAY_AMOUNT);
				break;
			case ANG_RADIAN:
				work_amount = DISPLAY_AMOUNT;
				break;
		}
	}
	DISPLAY_AMOUNT = (Trig_1[hyp_mode][data])(work_amount);	
/*
 * Now a cheat to help the weird case of COS 90 degrees not being 0!!!
 */
	if (DISPLAY_AMOUNT < POS_ZERO && DISPLAY_AMOUNT > NEG_ZERO)
		DISPLAY_AMOUNT=0;
	refresh_display = 1;
	UpdateDisplay();
}

void ExecATrig(FL_OBJECT *ob, long data)
{
	CALCAMNT 	work_amount;

	if (display_error)
		return;

	work_amount = (Trig_3[hyp_mode][data])(DISPLAY_AMOUNT);	
	if (errno == EDOM || errno == ERANGE)
		display_error = 1;
	else if (hyp_mode)
		DISPLAY_AMOUNT = work_amount;
	else {
		switch (angle_mode) {
			case ANG_DEGREE:
				DISPLAY_AMOUNT = RAD2DEG(work_amount);
				break;
			case ANG_GRADIENT:
				DISPLAY_AMOUNT = RAD2GRA(work_amount);
				break;
			case ANG_RADIAN:
				DISPLAY_AMOUNT = work_amount;
				break;
		}
	}
	refresh_display = 1;
	UpdateDisplay();
}

void EnterPercent(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	percent_mode = 1;
	EnterEqual(ob, data);
	percent_mode = 0;
}

void EnterLogr(FL_OBJECT *ob, long data)
{
	int button = temp_get_button_numb(ob);
	if (display_error)
		return;

	if (button > FL_SHORTCUT)
		button = (button > (FL_SHORTCUT + XK_F12)) ? 3 : 1;
	
	if (button < 3) {
		if (DISPLAY_AMOUNT <= 0)
			display_error = 1;
		else
			DISPLAY_AMOUNT = log10l(DISPLAY_AMOUNT);	
		refresh_display = 1;
		UpdateDisplay();
	} else {
		DISPLAY_AMOUNT = powl(10, DISPLAY_AMOUNT);
		refresh_display = 1;
		UpdateDisplay();
	}
	overlay_function = 0;
}

void EnterLogn(FL_OBJECT *ob, long data)
{
	int button = temp_get_button_numb(ob);
	if (display_error)
		return;

	if (button > FL_SHORTCUT)
		button = (button > (FL_SHORTCUT + XK_F12)) ? 3 : 1;
	

	if (button < 3) {
		if (DISPLAY_AMOUNT <= 0)
			display_error = 1;
		else
			DISPLAY_AMOUNT = logl(DISPLAY_AMOUNT);	
		refresh_display = 1;
		UpdateDisplay();
	} else {
		DISPLAY_AMOUNT = expl(DISPLAY_AMOUNT);	
		refresh_display = 1;
		UpdateDisplay();
	}
	overlay_function = 0;
}

void CycleAngle(FL_OBJECT *ob, long data) {

	switch (angle_mode)
	{
		case ANG_DEGREE:
			SetRad(ob, data);
			break;
		case ANG_GRADIENT:
			SetDeg(ob, data);
			break;
		case ANG_RADIAN:
			SetGra(ob, data);
			break;
	}
}

void SetDeg(FL_OBJECT *ob, long data) {
	angle_mode = ANG_DEGREE;
	fl_set_button(fd_calculator->DegSwitch, 1);
}

void SetGra(FL_OBJECT *ob, long data) {
	angle_mode = ANG_GRADIENT;
	fl_set_button(fd_calculator->GraSwitch, 1);
}

void SetRad(FL_OBJECT *ob, long data) {
	angle_mode = ANG_RADIAN;
	fl_set_button(fd_calculator->RadSwitch, 1);
}

void CycleBase(FL_OBJECT *ob, long data) {

	switch (current_base)
	{
		case 16:
			SetDec(ob, data);
			break;
		case 10:
			SetOct(ob, data);
			break;
		case 8:
			SetBin(ob, data);
			break;
		case 2:
			SetHex(ob, data);
			break;
	}
}

void SetHex(FL_OBJECT *ob, long data) {
	/*
	 * Set Hex Mode
	 */
	if (display_error)
		return;

	current_base = NB_HEX;
	display_size = HEX_SIZE;
	decimal_point = 0;
	input_limit = sizeof(long) * 2;
	fl_activate_object(fd_calculator->HexDigits);
	fl_set_object_lcol(fd_calculator->HexDigits, FL_BLACK);
	fl_activate_object(fd_calculator->BinDigits);
	fl_set_object_lcol(fd_calculator->BinDigits, FL_BLACK);
	fl_activate_object(fd_calculator->OctDigits);
	fl_set_object_lcol(fd_calculator->OctDigits, FL_BLACK);
	fl_activate_object(fd_calculator->DecDigits);
	fl_set_object_lcol(fd_calculator->DecDigits, FL_BLACK);
	fl_hide_object(fd_calculator->Decimal);
	fl_show_object(fd_calculator->Comma);
	fl_set_button(fd_calculator->CheckHex, 1);

	UpdateDisplay();
}

void SetOct(FL_OBJECT *ob, long data) {
	/*
	 * Set Oct Mode
	 */

	if (display_error)
		return;

	current_base = NB_OCTAL;
	display_size = OCT_SIZE;
	decimal_point = 0;
	input_limit = 11;
	fl_deactivate_object(fd_calculator->HexDigits);
	fl_set_object_lcol(fd_calculator->HexDigits, FL_INACTIVE);
	fl_deactivate_object(fd_calculator->DecDigits);
	fl_set_object_lcol(fd_calculator->DecDigits, FL_INACTIVE);

	fl_activate_object(fd_calculator->BinDigits);
	fl_set_object_lcol(fd_calculator->BinDigits, FL_BLACK);
	fl_activate_object(fd_calculator->OctDigits);
	fl_set_object_lcol(fd_calculator->OctDigits, FL_BLACK);
	fl_set_button(fd_calculator->OctSwitch, 1);
	fl_hide_object(fd_calculator->Decimal);
	fl_show_object(fd_calculator->Comma);

	UpdateDisplay();
}

void SetBin(FL_OBJECT *ob, long data) {
	/*
	 * Set Bin Mode
	 */

	if (display_error)
		return;

	current_base = NB_BINARY;
	display_size = BIN_SIZE;
	decimal_point = 0;
	input_limit = 32;
	fl_deactivate_object(fd_calculator->HexDigits);
	fl_set_object_lcol(fd_calculator->HexDigits, FL_INACTIVE);
	fl_deactivate_object(fd_calculator->OctDigits);
	fl_set_object_lcol(fd_calculator->OctDigits, FL_INACTIVE);
	fl_deactivate_object(fd_calculator->DecDigits);
	fl_set_object_lcol(fd_calculator->DecDigits, FL_INACTIVE);

	fl_activate_object(fd_calculator->BinDigits);
	fl_set_object_lcol(fd_calculator->BinDigits, FL_BLACK);
	fl_set_button(fd_calculator->BinSwitch, 1);
	fl_hide_object(fd_calculator->Decimal);
	fl_show_object(fd_calculator->Comma);

	UpdateDisplay();
}

void SetDec(FL_OBJECT *ob, long data) 
{
	/*
	 * Set Dec Mode
	 */

	if (display_error)
		return;

	current_base = NB_DECIMAL;
	display_size = DEC_SIZE;
	input_limit = 0;
	fl_deactivate_object(fd_calculator->HexDigits);
	fl_set_object_lcol(fd_calculator->HexDigits, FL_INACTIVE);

	fl_activate_object(fd_calculator->BinDigits);
	fl_set_object_lcol(fd_calculator->BinDigits, FL_BLACK);
	fl_activate_object(fd_calculator->OctDigits);
	fl_set_object_lcol(fd_calculator->OctDigits, FL_BLACK);
	fl_activate_object(fd_calculator->DecDigits);
	fl_set_object_lcol(fd_calculator->DecDigits, FL_BLACK);
	fl_set_button(fd_calculator->DecSwitch, 1);
	fl_show_object(fd_calculator->Decimal);
	fl_hide_object(fd_calculator->Comma);

	UpdateDisplay();
}

void Memory(FL_OBJECT *ob, long data)
{
	static CALCAMNT	memory_num = 0;
	int button;

	if (display_error)
		return;

	if (data == 0) {
		refresh_display = 1;
		memory_num = 0;
	} else if (data == 1) {
		DISPLAY_AMOUNT = memory_num;
		overlay_function = 0;
		refresh_display = 1;
		UpdateDisplay();
	} else if (data == 2) { 
	   	button = temp_get_button_numb(ob);
		
	  	if (button > FL_SHORTCUT)
			button = (button == FL_SHORTCUT + XK_Left) ? 3 : 1; 

		EnterEqual(ob, data);
		if (button == 1)
			memory_num += DISPLAY_AMOUNT;
		else
			memory_num -= DISPLAY_AMOUNT;
	} else if (data == 3) {
		memory_num = DISPLAY_AMOUNT;
		refresh_display = 1;
	}
}

void EnterEqual(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	overlay_function = 0;
	PushStack(&display_data);
	refresh_display = 1;
	old_display_index = 0;
	if (UpdateStack(0))
		UpdateDisplay();
	precedence_base = 0;
}

void Backspace(FL_OBJECT *ob, long data)
{
	int button;

	if (display_error) {
		ClearAll(ob, data);
	}
	else if (old_display_index > 0)
	{
		DISPLAY_AMOUNT = old_display[--old_display_index].amount;
		decimal_point  = old_display[old_display_index].decimal;
		UpdateDisplay();
	}
}

void Clear(FL_OBJECT *ob, long data)
{
	int button = temp_get_button_numb(ob);
	if (display_error) {
		ClearAll(ob, data);
	}
	else if (button == 1)
	{
		DISPLAY_AMOUNT = 0;
		old_display_index = 0;
		decimal_point = 0; 
		UpdateDisplay();
	}
	else if (old_display_index > 0)
	{
		Backspace(ob, data);
	}
	overlay_function = 0;
}

void ClearAll(FL_OBJECT *ob, long data)
{
	RefreshCalculator();
	overlay_function = 0;
	refresh_display = 1;
}

void ExitProgram(FL_OBJECT *ob, long data)
{
	if (data)
		PrintStack();
	fl_finish();
	exit(data);
}

char		display_str[DSP_SIZE+1];

void UpdateDisplay(void)
{
	CALCAMNT	boh_work_d;
	long 		boh_work, boh_work_h = 0;
	int		str_size;

	if (current_base != NB_DECIMAL) { 
		modfl(DISPLAY_AMOUNT, &boh_work_d);
		if (boh_work_d < LONG_MIN || boh_work_d > ULONG_MAX)
			display_error = 1;

	/*
     * We may be in that never-never land where boh numbers
	 * turn from positive to negative - if so then we do
     * just that, allowing boh negative numbers to be entered
     * as read (from dumps and the like!)
     */    
		else if (boh_work_d > LONG_MAX) 
			boh_work = DISPLAY_AMOUNT = 
				LONG_MIN+(boh_work_d-LONG_MAX-1);
		else 
			boh_work = DISPLAY_AMOUNT = boh_work_d;
	}

	if (!display_error) {

		if (current_base == NB_BINARY)
			str_size = cvb(display_str, 
			  	       boh_work, 
				       BOH_SIZE);
		else if (current_base == NB_OCTAL)
			str_size = sprintf(display_str, 
					   "%.0lo%lo", 
					   boh_work_h,boh_work);
		else if (current_base == NB_DECIMAL) {
			str_size = sprintf(display_str, 
					   "%.*Lg",
					   (DISPLAY_AMOUNT < 0 ? 13 : 14), 
					   DISPLAY_AMOUNT);
			sscanf(display_str, 
			       "%Lg", 
			       &DISPLAY_AMOUNT);
		} else if (current_base == NB_HEX)
			str_size = sprintf(display_str, 
					   "%.0lX%lX", 
					   boh_work_h,boh_work);
		else
			display_error = 1;
	}

	if (display_error || str_size > display_size || str_size < 0) { 
		display_error = 1;
		strcpy(display_str,"ERROR");
	}

	input_count = str_size;
	fl_set_object_color(fd_calculator->CalcDisplay, FL_TEXT_COL1, FL_TEXT_COL2);
	fl_set_object_lcolor(fd_calculator->CalcDisplay, FL_TEXT_LCOL);
	fl_set_object_label(fd_calculator->CalcDisplay, display_str);


	if (refresh_display)
		old_display_index = 0;
}

int cvb(char *out_str, long amount, int max_digits)
{
	/*
	 * A routine that converts a long int to 
	 * binary display format
	 */

	char		work_str[(sizeof(amount) * CHAR_BIT) + 1];
	int		work_char = 0, 
			lead_zero = 1, 
			lead_one = 1,
			lead_one_count = 0,
			work_size = sizeof(amount) * CHAR_BIT;
	unsigned long	bit_mask = (1 << ((sizeof(amount) * CHAR_BIT) - 1));

	while (bit_mask) {

		if (amount & bit_mask) {
			if (lead_one)
				lead_one_count++;
			lead_zero = 0;
			work_str[work_char++] = '1';
		} else {
			lead_one = 0;
			if (!lead_zero) 
				work_str[work_char++] = '0';
		}
		bit_mask >>= 1;
	}
	if (!work_char)
		work_str[work_char++] = '0';
	work_str[work_char] = '\0';

	if (work_char-lead_one_count < max_digits)
		return strlen(strcpy(out_str, 
				      &work_str[lead_one_count ? 
				      work_size - max_digits :
				      0]));
	else
	   return -1; 
}		

void cut_display(void)
{
  		fl_set_object_color(fd_calculator->CalcDisplay, FL_BLACK, FL_WHITE);
		fl_set_object_lcolor(fd_calculator->CalcDisplay, FL_WHITE);
		fl_stuff_clipboard(fd_calculator->CalcDisplay,FL_STRING, 
			 display_str, strlen(display_str), LoseSel);
}

void paste_display(void)
{
	if (fl_request_clipboard(fd_calculator->CalcDisplay,FL_STRING, GainSel)==0)
	{
		fl_sel_frozen = 1;
		fl_freeze_form(fd_calculator->calculator);
	}
}

void InsertConstant(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	DISPLAY_AMOUNT = constants[fl_get_menu(ob) - 1];
	overlay_function = 0;
	refresh_display = 1;
	UpdateDisplay();
}

void CutPasteDisplay(FL_OBJECT *ob, long data)
{
	if (display_error)
		return;

	if (data == 0)
		data = fl_get_menu(ob);

	if (data == 1)
		cut_display();
	else
		paste_display();
}


int LoseSel(FL_OBJECT *ob, long type)
{
	fl_set_object_color(fd_calculator->CalcDisplay, FL_TEXT_COL1, FL_TEXT_COL2);
	fl_set_object_lcolor(fd_calculator->CalcDisplay, FL_TEXT_LCOL);
	return 0;	
}

int GainSel(FL_OBJECT *ob, long type, const void *data, long size)
{
	int CheckBase(char *);
	CALCAMNT paste_amount;
	int		 base;
	long     paste_boh;
	
	if (size > 0)
	{
		/*
		 * Paste the data onto the display  
		 */
		base = CheckBase((char *)data);
		if (base == 0 && sscanf((char *)data, CALCFRMT, &paste_amount)==1)
		{
			DISPLAY_AMOUNT = paste_amount;
			refresh_display = 1;
			UpdateDisplay();
		} 
		else if (base == 2 && sscanf((char *)data, "%Li", &paste_boh)==1)
		{
			DISPLAY_AMOUNT = paste_boh;
			refresh_display = 1;
			UpdateDisplay();
		}
		else
		{
			/* Its binary - no implemented yet */
		}

	}

	if (fl_sel_frozen)
	{
		fl_unfreeze_form(fd_calculator->calculator);
		fl_sel_frozen = 0;	
	}
	return 0;
}

int CheckBase(char * str)
{
	/*
	 * Try to derive what type of number we have
	 * Return 1 for Binary, 2 for Octal or Hex, 0 for Decimal
	 */

	int j, base = 1;
	int	l_str = strlen(str);

	if (l_str == 0)
		return 0;

	if (l_str > 2 && str[0] == '0' && toupper(str[1]) == 'X')
		return 2;		/* Its Hex */

	/*
	 * At this point it could be decimal, octal or binary.
	 * If it does NOT have a leading zero then its assumed decimal
	 */

	if (str[0] != '0')
		return 0;		/* Its decimal */

	/*
	 * OK, this is a bit tougher. If it contains only binary digits
	 * before something that is not allowable, then its binary
	 * if it contains only octal digits then its octal
	 * otherwise it is decimal. HOWEVER note that '+-eE' are allowable
	 * and, if encountered, will determine it as decimal!
	 */

	for (j = 0; j < l_str; ++j)
	{
		if (isdigit(str[j]))
		{
			if (str[j] > '1')
				base = 2;
			if (str[j] > '7')
			{
				base = 0;
				break;
			}
		}
		else if (str[j] == '+' 
			 ||  str[j] == '-' 
			 ||  str[j] == '.' 
			 ||  toupper(str[j]) == 'E')
		{	
			base = 0;
			break;
		} 
		else
		{
			break;
		}
	}

	return base;
}

int UpdateStack(int run_precedence)
{
	item_contents 	new_item, *top_item , *top_function;
	CALCAMNT	left_op, right_op;
	int		op_function, return_value = 0;

	new_item.s_item_type = ITEM_AMOUNT;
	while ((top_function = TopTypeStack(ITEM_FUNCTION)) &&
	top_function->s_item_data.item_func_data.item_precedence >= 
	run_precedence) {

		return_value = 1;

		if ((top_item = PopStack())->s_item_type != ITEM_AMOUNT){
			printf("Stack processing error - right_op\n");
			ExitProgram(NULL, 1);
		}
		right_op = top_item->s_item_data.item_amount;

		if (!((top_item = PopStack()) && 
		top_item->s_item_type == ITEM_FUNCTION)) {
			printf("Stack processing error - function\n");
			ExitProgram(NULL, 1);
		}
		op_function = 
			top_item->s_item_data.item_func_data.item_function;	

		if (!((top_item = PopStack()) && 
		top_item->s_item_type == ITEM_AMOUNT)) {
			printf("Stack processing error - left_op\n");
			ExitProgram(NULL, 1);
		}
		left_op = top_item->s_item_data.item_amount;
	
		new_item.s_item_data.item_amount = 
			(Arith_ops[op_function])(left_op, right_op);
		PushStack(&new_item); 
			
	}	
	if (return_value && 
	    percent_mode && 
	    !display_error &&
	    Prcnt_ops[op_function] != NULL){
		new_item.s_item_data.item_amount =
			(Prcnt_ops[op_function])(left_op,
				right_op, 
				new_item.s_item_data.item_amount); 
		PushStack(&new_item);
	} 
	if (return_value)
		DISPLAY_AMOUNT = new_item.s_item_data.item_amount;

	return return_value;
}

int Exec01Function(int findex)
{
	CALCAMNT 		new_value;
	int				rc;
	char 			buf[BUF_SIZE], *args = NULL;

	/*
 	 * Zero and single argument functions are assumed to provide a new
 	 * DISPLAY_AMOUNT, possibly based upon the current display amount. 
	 * Although both types get passed the DISPLAY_AMOUNT, 0 arg functions
	 * should ignore it!!
	 */
	sprintf(buf, CALCFRMT, DISPLAY_AMOUNT);
	args = buf;

	rc = RunFunction(ext_funcs[findex].name, 1, &args, &new_value);

	if (rc == 0)
		DISPLAY_AMOUNT = new_value;

	return rc;	
}

int ExecFFunction(int findex)
{
printf("Running Fixed args\n");
	return 0;
}

int ExecVFunction(int findex)
{
	item_contents 	*top_item , *top_function;
	CALCAMNT		arg_vals[100], new_value;
	int				i=0, keep_going = 1, rc, argc;
	char 			**argv, *args;

	top_item = TopOfStack();

	/*
	 * If top-of-stack is a non-comma operation, we assume that the function is 
	 * going to operate only on the display value and, thus, call
	 * Exec01Function to manage the processing. 
	 * Otherwise we push the display value if the overlay_function flag
	 * is not set.
	 */
	if ((!top_item) || (top_item->s_item_type == ITEM_FUNCTION
	&& top_item->s_item_data.item_func_data.item_function != FUNC_COMMA))
		return Exec01Function(findex);

	if (!overlay_function)
		PushStack(&display_data);

	/*
	 * If the top item is a function, we replace it with
	 * our own function
	 */
	if ((top_item = TopOfStack())
	&& (top_item->s_item_type == ITEM_FUNCTION))
		PopStack();

	while ((top_item = TopOfStack()) && keep_going)
	{
		if ((top_item = PopStack())->s_item_type != ITEM_AMOUNT){
			printf("Stack processing error - right_op\n");
			ExitProgram(NULL, 1);
		}

		arg_vals[i++] = top_item->s_item_data.item_amount;

		if ((top_item = TopTypeStack(ITEM_FUNCTION)))
			if (top_item->s_item_data.item_func_data.item_function 
			== FUNC_COMMA 
			&& top_item->s_item_data.item_func_data.item_precedence 
			> precedence_base)
				PopStack();
			else				
				keep_going = 0;
			
	}	
	/*
	 * The arg values are now in arg_vals. Now build the parm list.
	 * Now we make the huge rash assumption that no value can possibly
	 * be bigger than the display size. Not really rash!
	 */

	argv = (char **)malloc((i*sizeof(char *)) + (i*(display_size+1))); 
	args = (char *)(argv+i);
	argc = i;

	for(i=0; i<argc; ++i)
	{
		argv[i] = args;
		sprintf(args, CALCFRMT, arg_vals[i]);
		args += (display_size+1);
	}

	rc = RunFunction(ext_funcs[findex].name, argc, argv, &new_value);

	if (rc == 0)
		DISPLAY_AMOUNT = new_value;

	free(argv);
	return rc;	
}			
	
int		RunFunction(char *name, int argc, char **argv, CALCAMNT *result)
{
	/*
	 * This does the technical stuff of setting up pipes, forking
	 * the new process, exec-ing perl with the required arguments
	 * and waiting on the results!
	 */

	char **	realargv, out_fd[BUF_SIZE];
	int	pipes[2], realargc, i, func_stat, rc, cpid;

	if (pipe(pipes) != 0)
		return -1;

	sprintf(out_fd, "%i", pipes[1]);
	realargc = argc + 4;		/* perl, script name, file desc, and NULL */
	realargv = (char **)calloc(realargc, sizeof(char *));

	realargv[0] = "perl";
	realargv[1] = name;
	realargv[2] = out_fd;

	for (i=3; argc; ++i)
		realargv[i] = argv[--argc];

	realargv[i] = NULL;

	if ((cpid = fork()) < 0)
		return -2;
	else if(cpid == 0)
	{
		execvp("perl", realargv);
		exit(-1);
	}
 
	/*
	 * We are the parent if we get here!!
	 */
	close(pipes[1]);
	waitpid(cpid, &func_stat, 0);
	if (WIFEXITED(func_stat))
		rc = WEXITSTATUS(func_stat);
	else
		rc = -1;

	if (rc == 0)
	{
		read(pipes[0], out_fd, BUF_SIZE);
		sscanf(out_fd, CALCFRMT, result);
	}

	close(pipes[2]);
	free(realargv);
	return rc;
}
		
 
CALCAMNT ExecOr(CALCAMNT left_op, CALCAMNT right_op)
{
	CALCAMNT	boh_work_d;
	long 		boh_work_l, boh_work_r;

	modfl(left_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_l = boh_work_d;
	modfl(right_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_r = boh_work_d;
	return (boh_work_l | boh_work_r);
}

CALCAMNT ExecXor(CALCAMNT left_op, CALCAMNT right_op)
{
	CALCAMNT	boh_work_d;
	long 		boh_work_l, boh_work_r;

	modfl(left_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_l = boh_work_d;
	modfl(right_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_r = boh_work_d;
	return (boh_work_l ^ boh_work_r);
}

CALCAMNT ExecAnd(CALCAMNT left_op, CALCAMNT right_op)
{
	CALCAMNT	boh_work_d;
	long 		boh_work_l, boh_work_r;

	modfl(left_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_l = boh_work_d;
	modfl(right_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_r = boh_work_d;
	return (boh_work_l & boh_work_r);
}

CALCAMNT ExecLsh(CALCAMNT left_op, CALCAMNT right_op)
{
	CALCAMNT	boh_work_d;
	long 		boh_work_l, boh_work_r;

	modfl(left_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_l = boh_work_d;
	modfl(right_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_r = boh_work_d;
	return (boh_work_l << boh_work_r);
}

CALCAMNT ExecRsh(CALCAMNT left_op, CALCAMNT right_op)
{
	CALCAMNT	boh_work_d;
	long 		boh_work_l, boh_work_r;

	modfl(left_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_l = boh_work_d;
	modfl(right_op, &boh_work_d);
	if (fabsl(boh_work_d) > LONG_MAX) {
		display_error = 1;    
		return 0;
	}
	boh_work_r = boh_work_d;
	return (boh_work_l >> boh_work_r);
}

CALCAMNT ExecAdd(CALCAMNT left_op, CALCAMNT right_op)
{
	return left_op + right_op;
}

CALCAMNT ExecSubtract(CALCAMNT left_op, CALCAMNT right_op)
{
	return left_op - right_op;
}

CALCAMNT ExecMultiply(CALCAMNT left_op, CALCAMNT right_op)
{
	return left_op * right_op;
}

CALCAMNT ExecComma(CALCAMNT left_op, CALCAMNT right_op)
{
	/*
	 * If we attempt to roll the stack with comma operations but no
	 * functions, then we interpret the comma as a replacement operation
	 * replacing the first operand with the second operand.
	 */
	return right_op;
}

CALCAMNT ExecDivide(CALCAMNT left_op, CALCAMNT right_op)
{
	if (right_op == 0) {
		display_error = 1;
		return 0;
	} else 
		return left_op / right_op;
}

CALCAMNT ExecMod(CALCAMNT left_op, CALCAMNT right_op)
{
	if (right_op == 0) {
		display_error = 1;
		return 0;
	} else 
		return fmodl(left_op, right_op);
}

CALCAMNT ExecIntDiv(CALCAMNT left_op, CALCAMNT right_op)
{
	if (right_op == 0) {
		display_error = 1;
		return 0;
	} else {
		modfl(left_op / right_op, &left_op); 
		return left_op;
	}
}

CALCAMNT ExecPower(CALCAMNT left_op, CALCAMNT right_op)
{
	if (right_op == 0)
		return 1;
	if (left_op < 0 && isoddint(1/right_op)) 
		left_op = -1 * powl((-1 * left_op), right_op);
	else
		left_op = powl(left_op, right_op);
	if (errno == EDOM || errno == ERANGE) {
		display_error = 1;
		return 0;
	} else 
		return left_op;
}

CALCAMNT ExecPwrRoot(CALCAMNT left_op, CALCAMNT right_op)
{
	if (right_op == 0) {
		display_error = 1;
		return 0;
	}
	if (left_op < 0 && isoddint(right_op)) 
		left_op = -1 * powl((-1 * left_op), 1/right_op);
	else
		left_op = powl(left_op, 1/right_op);
	if (errno == EDOM || errno == ERANGE) {
		display_error = 1;
		return 0;
	} else 
		return left_op;
}

int isoddint(CALCAMNT input) 
{
	CALCAMNT	dummy;
	/*
	 * Routine to check if CALCAMNT is an Odd integer
	 */
	return (modfl(input, &dummy) == 0.0 &&
		modfl(input/2, &dummy) == 0.5);
}
	
CALCAMNT ExecAddSubP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result)
{
	if (result == 0) {
		display_error = 1;
		return 0;
	} else 
		return (result * 100) / right_op;
}

CALCAMNT ExecMultiplyP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result)
{
	return (result / 100);
}

CALCAMNT ExecDivideP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result)
{
	return (result * 100);
}

CALCAMNT ExecPowerP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result)
{
	return ExecPower(left_op, (right_op / 100));
}

CALCAMNT ExecPwrRootP(CALCAMNT left_op, CALCAMNT right_op, CALCAMNT result)
{
	if (right_op == 0) {
		display_error = 1;
		return 0;
	} else 
		return ExecPower(left_op, (100 / right_op));
}

void CalcHelp(FL_OBJECT *ob, long data)
{
	   fl_show_form(fd_calchelp->calchelp,FL_PLACE_FREE,FL_TRANSIENT,"Calculator Help");
}

/******************************************************************************
*******************************************************************************
*******************************************************************************
 *
 * Stack Data
 *
 * These typedefs define the stack data items.
 *
 */

typedef struct stack_item	*stack_ptr;
typedef struct stack_item {

	/* Contents of an item on the input stack */

	stack_ptr	prior_item;		/* Pointer to prior item */
	stack_ptr	prior_type;		/* Pointer to prior type */
	item_contents	item_value;		/* The value of the item */

} stack_item;					/* all called stack_item */

stack_ptr	top_of_stack = NULL;		/* Current stack top     */
stack_ptr	top_type_stack[2] = { NULL, NULL };	
						/* Current top type      */

void 	  UnAllocStackItem (stack_ptr return_item);
stack_ptr AllocStackItem   (void);

/*
 * Stack Operations
 *
 *
 */

void PushStack(item_contents *add_item) 
{
	/*
	 * Add an item to the stack
	 */

	stack_ptr new_item = top_of_stack;
	
	/*printf("Start Push: ");
	PrintStack();*/

	if (!(new_item && 
	new_item->item_value.s_item_type == add_item->s_item_type)) {

		new_item = AllocStackItem();	/* Get a new item    */

		/*
		 * Chain new item to existing stacks 
		 */

		new_item->prior_item = top_of_stack;
		top_of_stack	     = new_item;
		new_item->prior_type = top_type_stack[add_item->s_item_type];
		top_type_stack[add_item->s_item_type] = new_item;
	}

	new_item->item_value  = *add_item;	/* assign contents*/

	/*printf("End Push: ");
	PrintStack();*/
}

item_contents *PopStack(void)
{
	/*
	 * Remove and return the top item in the stack
	 */

	static item_contents return_item;
	item_contents *return_item_ptr = NULL;
	stack_ptr return_stack_ptr;

	/*printf("Start Pop: ");
	PrintStack();*/

	if (return_stack_ptr = top_of_stack) {
		return_item = top_of_stack->item_value;

		top_type_stack[return_item.s_item_type]
			     = top_of_stack->prior_type;
		top_of_stack = top_of_stack->prior_item;

		UnAllocStackItem(return_stack_ptr);
		return_item_ptr = &return_item;
	}

	/*printf("End Pop: ");
	PrintStack();*/
	return return_item_ptr;
}

item_contents *TopOfStack(void)
{
	/*
	 * Return the top item in the stack without removing
	 */

	item_contents *return_item_ptr = NULL;

	if (top_of_stack) {
		return_item_ptr = &top_of_stack->item_value;
	}

	return return_item_ptr;
}

item_contents *TopTypeStack(item_type rqstd_type)
{
	/*
	 * Return the top item in the stack without removing
	 */

	item_contents *return_item_ptr = NULL;

	if (top_type_stack[rqstd_type]) {
		return_item_ptr = &top_type_stack[rqstd_type]->item_value;
	}

	return return_item_ptr;
}

void PrintStack()
{
	/*
	 * A diagnostic routine that prints the stack
	 */

	stack_ptr next_item = top_of_stack;
	printf("Start of stack print:\n");
	
	printf("\tTop Of Stack: %p\n\tTop Amount: %p\n\tTop Function: %p\n", 
		top_of_stack, 
		top_type_stack[ITEM_AMOUNT], 
		top_type_stack[ITEM_FUNCTION]); 

	while (next_item) {

		printf("%9p Prior-> %9p\tPriorType-> %9p\t", 
			next_item, 
			next_item->prior_item, 
			next_item->prior_type);
		if (next_item->item_value.s_item_type == ITEM_FUNCTION) {
			printf ("FUNCTION:\t%i\n", 
			    	next_item->item_value.s_item_data);
		} else if (next_item->item_value.s_item_type == ITEM_AMOUNT) {
			printf ("AMOUNT:\t%Lf\n", 
			    	next_item->item_value.s_item_data);
		} else {
			printf ("UNKNOWN TYPE:\t%i\n", 
			    	next_item->item_value.s_item_type);
		}
		next_item = next_item->prior_item;
	}
	printf("End of stack print:\n\n");
}

/*
 * Stack storage management Data and Functions
 */

stack_item	process_stack[STACK_SIZE];
int		stack_next;
int		stack_last;

void InitStack (void) {
	
	stack_next = 0;
	stack_last = STACK_SIZE - 1;
	top_of_stack = top_type_stack[0] = top_type_stack[1] = NULL;
}

stack_ptr AllocStackItem (void) {

	if (stack_next <= stack_last) {

		process_stack[stack_next].prior_item = NULL;
		process_stack[stack_next].prior_type = NULL;
		return &process_stack[stack_next++];
	}

	printf("Stack Overflow!!!!!!\n");
	fl_finish();
	exit(1);
}

void UnAllocStackItem (stack_ptr return_item) {

	if (return_item != &process_stack[--stack_next]) {
	
		printf("Stack Error!!!!!!\n");
		fl_finish();
		exit(1);

	}

}

int	 button1_down = 0;
Time button1_at	  = 0;
int  button2_down = 0;

int raw_calc_callback(FL_FORM *form, void *xevent)
{
	/*
	 * Some dirty tricks involving mouse clicks in the display, thus
	 * allowing cut and paste by mouse!
	 */

	FL_Coord mx, my, dx, dy, dw, dh;
	unsigned keymask;
	int		 in_display;

	fl_get_form_mouse(form, &mx, &my, &keymask);
	fl_get_object_geometry(fd_calculator->CalcDisplay, &dx, &dy, &dw, &dh);

	if (!((mx >= dx && mx <= (dx + dw)) && (my >= dy && my <= (dy + dh))))
	{
	/*
	 * Circumventing a bug in Xforms! we save the button number now!!
	 */
	    temp_button_numb = ((XEvent *)xevent)->xbutton.button;
		button1_down = 0;
		button2_down = 0;
		button1_at   = 0;
		return 0;
	}
	
	switch (((XEvent *)xevent)->xbutton.button)
	{
		case Button1:
			if (((XEvent *)xevent)->xbutton.type == ButtonRelease)
			{
				if(button1_down == 2)
				{	
					cut_display();
					button1_down = 0;
					button1_at   = 0;
				}
			}
			else
			{
				if (button1_down 
				&& (((XEvent *)xevent)->xbutton.time - button1_at) 
				< FL_CLICK_TIMEOUT)
					button1_down = 2;
				else
				{ 
					button1_down = 1;
					button1_at   = ((XEvent *)xevent)->xbutton.time;
				}
			}
			break;
		case Button2:
		case Button3:
			if (((XEvent *)xevent)->xbutton.type == ButtonRelease)
			{
				if (button2_down)
					paste_display();
				button2_down = 0;
			}
			else 
			{
				button2_down = 1;
				button1_down = 0;
				button1_at   = 0;
			}
			break;
		default:
			button2_down = 0;
			button1_down = 0;
			button1_at   = 0;
			break;
	}

	return FL_PREEMPT;
} 
